import importlib
import json
import os
import pandas as pd
import shutil


class ModelExport:
    def __init__(self):
        self._FW_VERSION_DICT_ = {
            "sklearn": "0.24.2",
            "tensorflow": "2.2.0",
            "pytorch": "1.6.0"
        }
        self._DATA_TYPE_LIST_ = ["Tabular", "Image", "Text"]
        self._MODEL_TYPE_LIST_ = ["sklearn", "tensorflow", "torch"]
        self._ESTM_TYPE_LIST_ = ["classifier", "others", "regressor"]
        self._BASE_PATH_ = "/home/work"
        self._base_data_dir_ = os.path.join(self._BASE_PATH_, "dataset")
        self._schema_path_ = os.path.join(self._base_data_dir_, ".columns.json")

    def _validate_input(self, value, valid_values, name):
        if value not in valid_values:
            raise ValueError(f"{value}은(는) 유효하지 않은 {name}입니다. 가능한 값은 {valid_values} 입니다.")

    def modelValidationForAiruntime(self, model_type=None, data_type=None, estimator_type=None, sample_dataset=None, target_col=None):
        if any(param is None for param in [model_type, data_type, estimator_type]) \
                or sample_dataset is None or sample_dataset.empty:
            raise ValueError("model_type, data_type, sample_dataset, estimator_type은 반드시 입력되어야 합니다.")

        self._validate_input(model_type, self._MODEL_TYPE_LIST_, "모델 유형")
        self._validate_input(data_type, self._DATA_TYPE_LIST_, "데이터 유형")
        self._validate_input(estimator_type, self._ESTM_TYPE_LIST_, "분석 유형")

        if estimator_type == 'regressor' and target_col is None:
            raise ValueError("모델이 regressor일 경우 target_col 입력이 필요합니다.")

        _DEFAULT_TARGET_COL_ = target_col if estimator_type == 'regressor' else "variety"  # classification

        # datapath none -> _dataset 유무 무관
        if not os.path.exists(self._schema_path_):
            self._write_schema(sample_dataset, estimator_type=estimator_type)

        model_dir = os.path.join(
            "file://", self._BASE_PATH_, "airuntime", model_type, "models")

        _data_info = os.path.join(
            self._BASE_PATH_, "airuntime", model_type, "data-info.json",
        )
        _data_schema = os.path.join(
            self._BASE_PATH_, "airuntime", model_type, "columns.json",
        )
        if os.path.exists(_data_info):
            os.remove(_data_info)
        if os.path.exists(_data_schema):
            os.remove(_data_schema)

        try:
            with open(self._schema_path_, "r") as f:
                schema_info = json.load(f)
            with open(self._schema_path_, "w") as f:
                if _DEFAULT_TARGET_COL_:
                    schema_info.update({
                        "targetColumn": _DEFAULT_TARGET_COL_,
                        "estimatorType": estimator_type,
                    })
                json.dump(schema_info, f)
        except FileNotFoundError:
            print("Dataset을 선택하여야 합니다.")


        _label_encoder_path = os.path.join(model_dir, "labelEncoder.joblib")
        if estimator_type != "classifier" and os.path.exists(_label_encoder_path):
            os.remove(_label_encoder_path)

        data_info_src = os.path.join("file://", self._BASE_PATH_, self._schema_path_)
        data_info_dst = os.path.join("file://", _data_schema)
        shutil.copy(data_info_src, data_info_dst)

        self._write_model_flag(model_type, data_type)

    def _write_model_flag(self, model_type: str = "sklearn", data_type: str = "Tabular"):
        fw_name = model_type
        fw_module = importlib.import_module(fw_name)
        try:
            fw_version = fw_module.__version__ if fw_module.__version__ else self._FW_VERSION_DICT_.get(fw_name)
        except AttributeError:
            fw_version = self._FW_VERSION_DICT_.get(fw_name)
        model_flag = f"""# This file is a SYSTEM-GENERATED file. DO NOT EDIT THIS!!!

MODEL_TYPE={model_type}
FW={fw_name}
FW_VERSION={fw_version}
DATA_TYPE={data_type}
MODEL_TEST=/home/work/airuntime/{model_type}/test_server.py
"""
        # print(model_flag)
        # print(data_type.lower())
        with open(self._schema_path_, "r") as f:
            custom_schema = json.load(f)
        custom_name = custom_schema.get("filename")
        if data_type.lower() == "tabular":
            try:
                data_info, data_path = self._load_schema()
                data_path = data_path.split(self._BASE_PATH_ + os.path.sep)[-1]
                data_name = os.path.basename(data_path)
                if data_name == custom_name:
                    model_flag += "PIPELINE_SCHEMA={pipeline_schema}\n".format(
                        pipeline_schema=json.dumps(data_info["schema"]))
                else:
                    model_flag += "PIPELINE_SCHEMA={pipeline_schema}\n".format(pipeline_schema="")
            except FileNotFoundError:
                data_path = custom_name
                model_flag += "PIPELINE_SCHEMA={pipeline_schema}\n".format(pipeline_schema="")
            model_flag += "SOURCE_URL={data_path}\n".format(data_path=data_path)
        print(model_flag)
        with open(os.path.join(self._BASE_PATH_, ".MODEL_FLAG"), "w") as f:
            f.write(model_flag)

    def _load_schema(self):
        try:
            with open(os.path.join(self._BASE_PATH_, "dataset", "data-info.json"), "r") as f:
                data_info = json.load(f)[0]
                data_path = os.path.join(self._BASE_PATH_, 'dataset', data_info.get('filename', None))
                if not data_path:
                    raise ValueError("파이프라인 데이터 정보가 잘못되었습니다.")
                return data_info, data_path
        except FileNotFoundError as fe:
            raise FileNotFoundError("파이프라인 데이터 정보를 찾을 수 없습니다.")
        except IndexError as ie:
            raise IndexError("파이프라인 데이터 리스트가 잘못되었습니다.")
        except Exception as e:
            raise e

    def _write_schema(self, _dataset: pd.DataFrame, data_path=None, estimator_type="classifier"):
        os.makedirs(os.path.dirname(os.path.abspath(self._schema_path_)), exist_ok=True)
        if data_path is None:
            data_path = ".dummy.csv" # 추후 data_path도 user에게 받을 경우 사용
            _dataset.to_csv(os.path.join(self._BASE_PATH_, data_path), index=False)
        with open(self._schema_path_, "w") as f:
            json.dump({
                "filename": os.path.basename(data_path),
                "columns": _dataset.dtypes.index.tolist(),
                "datatypes": [d.name for d in _dataset.dtypes],
                "estimatorType": estimator_type,
            }, f)


from collections.abc import Mapping
from csv import DictReader
import json
from Accuinsight.modeler.core.LcConst.LcConst import RUN_RESULT_PATH, RUN_MODEL_VISUAL_CSV_PATH, RUN_MODEL_JSON_PATH, \
    SELECTED_METRICS, RUN_OBJ_OPT_INFO
from Accuinsight.modeler.protos.life_cycle_pb2 import LcMetric, LcParam
from Accuinsight.modeler.core.Run.ParseRun import ParserVisualJaon, Parse_Helper

# get parameter path with run type
# get metric path with run type

# parse metrics
def _parse_selected_metrics(run_meta):
    # read ~/runs/results-XGBClassifier/model-info-json/*.json
    # get "selected_metrics" field
    # return metrics

    grpc_metrics = []

    # selected_metrics
    selected_metrics = run_meta[SELECTED_METRICS]
    metric_keys = selected_metrics.keys()

    result_dict = {'metrics': []}

    count = 0
    timestamp = []
    steps = []
    values = []
    for key in metric_keys:
        values.append(str(selected_metrics[key]))
        timestamp.append(str(count))
        steps.append(key)
        count += 1

    result_dict['metrics'].append({
                                   'values': values,
                                   'timestamp': timestamp,
                                   'steps': steps})

    metrics = result_dict['metrics']

    for item in metrics:
        metric = LcMetric(
                key=SELECTED_METRICS,
                values=item['values'],
                timestamp=item['timestamp'],
                steps=item['steps']
            )
        grpc_metrics.append(metric)

    return grpc_metrics


def parse_metric(run_meta):
    # parse 'for-visual-csv/keras-history-n.csv'
    # metric_data = load_metric(run_meta)
    # metrics = metric_data['metrics']

    grpc_metrics = []

    selected_metrics = _parse_selected_metrics(run_meta=run_meta)
    grpc_metrics = selected_metrics
    return grpc_metrics


def _parse_optimizer_info(param_data):
    grpc_params = []

    optimizer_info = param_data
    if isinstance(optimizer_info, Mapping):
        for k, v in optimizer_info.items():
            optimizer = LcParam(
                key='type_of_optimizer',
                value=str(k)
            )
            grpc_params.append(optimizer)
            if isinstance(v, Mapping):
                for k, v in v.items():
                    param = LcParam(
                        key=k,
                        value=str(v)
                    )
                    grpc_params.append(param)
    return grpc_params


def parse_parameter(run_meta):
    # parse 'model-info-json/keras-n.json'
    param_data = run_meta[RUN_OBJ_OPT_INFO]
    dummy_data = {
        'data_version': 'https://refined-public-raw-d-curation.s3.ap-northeast-2.amazonaws.com/svdata_test/zipCode_sigunCode.csv',
        'model_description': 'keras test',
        'logging_time': '2020-05-06 10:30:50',
        'run_id': '9FCEF776-D830-4353-8961-144491DC03CC',
        'model_type': 'keras_nn',
        'optimizer_info': {
            'RMSprop': {
                'learning_rate': 9.999999747378752e-05,
                'rho': 0.8999999761581421,
                'decay': 0.0,
                'epsilon': 1e-07
            }
        },
        'time_delta': '0:00:14.912202'
    }

    return _parse_optimizer_info(param_data)


def parse_run_result(run_info_json):
    result_dict = {'metrics': None, 'params': None, 'visual': None}

    # parse 'for-visual-csv/keras-history-n.csv'
    metric_evl_data = parse_metric(run_meta=run_info_json)

    # parse 'model-info-json/keras-n.json'
    parameter_evl_data = parse_parameter(run_meta=run_info_json)
    # parse 'for-visual-json/keras-visual-n.json'
    visual_evl_data = ParserVisualJaon.parse_run_visual(run_mata=run_info_json)

    result_dict['metrics'] = metric_evl_data
    result_dict['params'] = parameter_evl_data
    result_dict['visual'] = visual_evl_data
    return result_dict


from Accuinsight.modeler.entities._modeler_object import _ModelerObject


class Run(_ModelerObject):
    """
    Run object.
    """

    def __init__(self, run_id):
        self._run_id = run_id

    @property
    def run_id(self):
        """
        The run id that is returned by back-end server.

        """
        return self._run_id


import logging
import re
import os
import json

logging.basicConfig(level=logging.DEBUG)


class AbcNotifier:
    def completed_message(self, **kwargs):
        pass


def load_model(run_name):
    best_model_path_dict = dict()

    model_type = run_name.split('-')[0]
    t = run_name.split('-')[1].split('_')[0]
    filename = model_type + '-' + t[0:8] + '-' + t[8:12] + '-' + t[12:16] + '-' + t[16:20] + '-' +  t[20:]
    regex = re.compile(filename)

    common_path = '/home/work/runs/best-model'

    for subdir, dirs, files in os.walk(common_path):
        dirs[:] = []  # 하위 디렉토리를 탐색하지 않도록 dirs 리스트를 비움
        for filename in files:
            if regex.search(filename):
                if filename.endswith(".joblib"):
                    best_model_path_dict['joblib'] = os.path.join(common_path, filename)
                elif filename.endswith(".json"):
                    best_model_path_dict['json'] = os.path.join(common_path, filename)
                elif filename.endswith(".h5"):
                    best_model_path_dict['h5'] = os.path.join(common_path, filename)

    # load model trained using keras
    if model_type == 'keras':
        from keras.models import model_from_json
        # load json and create model
        json_file = open(best_model_path_dict['json'], 'r')
        loaded_model_json = json_file.read()
        json_file.close()
        loaded_model = model_from_json(loaded_model_json)

        # load weights into new model
        loaded_model.load_weights(best_model_path_dict['h5'])

    # load model trained using tensorflow
    elif model_type == 'tf.keras':
        from tensorflow.keras.models import model_from_json
        # load json and create model
        json_file = open(best_model_path_dict['json'], 'r')
        loaded_model_json = json_file.read()
        json_file.close()
        loaded_model = model_from_json(loaded_model_json)

        # load weights into new model
        loaded_model.load_weights(best_model_path_dict['h5'])

    else:
        import joblib
        json_file = open(best_model_path_dict['json'], 'r')
        loaded_model_json = json_file.read()
        json_file.close()
        loaded_model = joblib.load(best_model_path_dict['joblib'])

    # 파라미터 추가
    # 실제로 experiment 모델을 생성하는 것이 아니라 평가해서 모델러 백엔드로 동일 모델에 대한 평가를 식별하기 위해 추가.
    loaded_model_dict = json.loads(loaded_model_json)

    loaded_model_dict["run_name"] = run_name
    update_loaded_model_json = json.dumps(loaded_model_dict)

    return loaded_model, update_loaded_model_json


def create_metrics_dict(key, values, steps=None):
    if key == "confusion_matrix":
        steps = list(map(str, range(len(values[0]))))
    elif key == "true_y" or key == "predicted_y":
        steps = list(map(str, range(len(values))))
    elif steps is None:
        raise ValueError("Steps parameter must be provided for metrics other than 'confusion_matrix', 'true_y', and 'predicted_y'")

    metrics_dict = {
        "key": key,
        "values": values,
        "timestamp": list(map(str, range(len(steps)))),
        "steps": steps
    }
    return metrics_dict


import base64
import time
import logging
import json

import requests

from Accuinsight.__about__ import __version__
from Accuinsight.modeler.utils.string_utils import strip_suffix

RESOURCE_DOES_NOT_EXIST = 'RESOURCE_DOES_NOT_EXIST'

_logger = logging.getLogger(__name__)

_DEFAULT_HEADERS = {
    'User-Agent': 'modeler-python-client/%s' % __version__
}


def http_request(host_creds, endpoint, retries=3, retry_interval=3,
                 max_rate_limit_interval=60, **kwargs):
    """
    Makes an HTTP request with the specified method to the specified hostname/endpoint. Ratelimit
    error code (429) will be retried with an exponential back off (1, 2, 4, ... seconds) for at most
    `max_rate_limit_interval` seconds.  Internal errors (500s) will be retried up to `retries` times
    , waiting `retry_interval` seconds between successive retries. Parses the API response
    (assumed to be JSON) into a Python object and returns it.

    :param host_creds (object containing)
        hostname and optional authentication.
    :return: Parsed API response
    """
    hostname = host_creds.host
    auth_str = None
    if host_creds.username and host_creds.password:
        basic_auth_str = ("%s:%s" % (host_creds.username, host_creds.password)).encode("utf-8")
        auth_str = "Basic " + base64.standard_b64encode(basic_auth_str).decode("utf-8")
    elif host_creds.token:
        auth_str = "Bearer %s" % host_creds.token

    headers = dict(_DEFAULT_HEADERS)
    headers['Content-Type'] = 'application/json'  # Add Content-Type header

    if auth_str:
        headers['Authorization'] = auth_str

    verify = not host_creds.ignore_tls_verification

    def request_with_ratelimit_retries(max_rate_limit_interval, **kwargs):
        response = requests.request(**kwargs)
        time_left = max_rate_limit_interval
        sleep = 1
        while response.status_code == 429 and time_left > 0:
            _logger.warning(
                "API request to {path} returned status code 429 (Rate limit exceeded). "
                "Retrying in %d seconds. "
                "Will continue to retry 429s for up to %d seconds.",
                sleep, time_left)
            time.sleep(sleep)
            time_left -= sleep
            response = requests.request(**kwargs)
            sleep = min(time_left, sleep*2)  # sleep for 1, 2, 4, ... seconds;
        return response

    cleaned_hostname = strip_suffix(hostname, '/')
    url = "%s%s" % (cleaned_hostname, endpoint)
    for i in range(retries):
        # response = request_with_ratelimit_retries(max_rate_limit_interval,
        #                                           url=url, headers=headers, verify=verify, **kwargs)
        response = request_with_ratelimit_retries(max_rate_limit_interval,
                                                  url=url, headers=headers, **kwargs)
        if response.status_code >= 200 and response.status_code < 500:
            return response
        else:
            _logger.error(
                "API request to %s failed with code %s != 200, retrying up to %s more times. "
                "API response body: %s",
                url, response.status_code, retries - i - 1, response.text)
            time.sleep(retry_interval)
    raise ValueError("API request to %s failed to return code 200 after %s tries" %
                          (url, retries))


def _can_parse_as_json(string):
    try:
        json.loads(string)
        return True
    except Exception:  # pylint: disable=broad-except
        return False

def verify_rest_response(response, endpoint):
    """Verify the return code and raise exception if the request was not successful."""
    if response.status_code != 200:
        if _can_parse_as_json(response.text):
            # raise RestException(json.loads(response.text))
            raise ValueError("RestException: '%s'", response.text)
        else:
            base_msg = "API request to endpoint %s failed with error code " \
                       "%s != 200" % (endpoint, response.status_code)
            raise ValueError("%s. Response body: '%s'" % (base_msg, response.text))
    return response


import json
import logging

from Accuinsight.modeler.store.tracking.lc_abstract_store import AbstractStore as LcAbstractStore
from Accuinsight.modeler.utils.os_getenv import get_os_env
from Accuinsight.modeler.core.Run.ParseRun import BaseParser
from Accuinsight.modeler.core.Run.RunInfo.RunInfo import dict_to_list
from Accuinsight.modeler.core.LcConst.LcConst import SOURCE_FILE_GIT_META, PYTHON_DEPENDENCY, SOURCE_FILE_GIT_REPO, \
    SOURCE_FILE_GIT_COMMIT_ID, RUN_RESULT_PATH, RUN_MODEL_SHAP_JSON_PATH
from Accuinsight.modeler.core.LcConst import LcConst
from Accuinsight.modeler.utils.rest_utils_runsandbox import http_request, verify_rest_response


def _set_run_info_dict(current_run_meta, user_sso_id):
    model_path = ''
    json_path = ''

    if current_run_meta[LcConst.RUN_OBJ_BEST_MODEL_H5_PATH] is not None:
        model_path = current_run_meta[LcConst.RUN_OBJ_BEST_MODEL_H5_PATH]

    if current_run_meta[LcConst.RUN_OBJ_BEST_MODEL_JOBLIB_PATH] is not None:
        model_path = current_run_meta[LcConst.RUN_OBJ_BEST_MODEL_JOBLIB_PATH]

    if current_run_meta[LcConst.RUN_OBJ_BEST_MODEL_JSON_PATH] is not None:
        json_path = current_run_meta[LcConst.RUN_OBJ_BEST_MODEL_JSON_PATH]

    run_info = {
        "userId": user_sso_id,
        "runNickName": current_run_meta[LcConst.RUN_OBJ_NICK_NAME],
        "creaDt": str(current_run_meta[LcConst.START_TIME]),
        "endDt": str(current_run_meta[LcConst.END_TIME]),
        "duration": str(current_run_meta[LcConst.DELTA_TIME]),
        "afLoc": "",
        "status": "LcFINISHED",
        "path": current_run_meta[LcConst.RUN_OBJ_MODEL_FILE_PATH],
        "modelPath": model_path,
        "jsonPath": json_path,
        "name": current_run_meta[LcConst.RUN_OBJ_NAME],
        "note": "",
        "isEvaluation": current_run_meta[LcConst.RUN_OBJ_IS_EVALUATION]
    }

    return run_info


class RestStore(LcAbstractStore):
    """
    Client for a remote tracking server accessed via REST API calls
    """

    def __init__(self, get_host_creds):
        super(RestStore, self).__init__()
        self.get_host_creds = get_host_creds

    @property
    def host_creds(self):
        return self.get_host_creds

    def _call_endpoint(self, json_body, current_run_meta):

        # '/project/{project}/workspace/{workspaceId}/experiment/{experimentId}/run'
        env_value = get_os_env()
        endpoint = '/project/{project}/workspace/{workspaceId}/experiment/{experimentId}/run'
        endpoint = endpoint.replace('{project}', str(env_value[LcConst.ENV_PROJECT_ID]))
        endpoint = endpoint.replace('{workspaceId}', str(env_value[LcConst.ENV_WORKSPACE_ID]))
        endpoint = endpoint.replace('{experimentId}', str(env_value[LcConst.ENV_EXPERIMENT_ID]))
        if json_body:
            data = json.loads(json_body)
            data = self._make_metrics_data(data)
            data = self._make_parameter_data(data)
            data = self._make_feature_data(data, current_run_meta)
        param = json.dumps(data)

        try:
            response = http_request(host_creds=self.host_creds, endpoint=endpoint, method='POST', data=param)
            response = verify_rest_response(response, endpoint).text

        except Exception as e:
            logging.error("Modeler API server connection failed", e)
            response = None

        return response

    def lc_create_run(self, current_run_meta):
        """
        Create a run under the specified experiment ID.

        :return: The ID of the created Run object
        """

        # read workspace environment
        # project_id, workspace_id, experiment_id and user_id
        env_value = get_os_env()
        project_id = str(env_value[LcConst.ENV_PROJECT_ID])
        workspace_id = str(env_value[LcConst.ENV_WORKSPACE_ID])
        experiment_id = str(env_value[LcConst.ENV_EXPERIMENT_ID])
        user_id = str(env_value[LcConst.ENV_USER_SSO_ID])  # owner id

        if current_run_meta[LcConst.RUN_OBJ_IS_EVALUATION]=="false":
            print("training model.")
            with open(current_run_meta[LcConst.RUN_RESULT_PATH][LcConst.RUN_MODEL_JSON_PATH]) as run_json_file:
                json_data = json.load(run_json_file)

                try:
                    user_id = json_data['user_id']
                except KeyError:
                    pass
        nick_name = current_run_meta[LcConst.NICK_NAME]
        run_proto = _set_run_info_dict(current_run_meta, user_id)

        git_meta = {'filename': '', 'repo': '', 'commit': ''}
        git_meta_data = {
            "url": git_meta[SOURCE_FILE_GIT_REPO],
            "commit": git_meta[SOURCE_FILE_GIT_COMMIT_ID]
        }

        if SOURCE_FILE_GIT_META in current_run_meta:
            git_meta = current_run_meta[SOURCE_FILE_GIT_META]
            if SOURCE_FILE_GIT_REPO in git_meta:
                git_meta_data = {
                    "url": git_meta[SOURCE_FILE_GIT_REPO],
                    "commit": git_meta[SOURCE_FILE_GIT_COMMIT_ID]
                }

        py_depen = {'data': ''}
        if PYTHON_DEPENDENCY in current_run_meta:
            pdep = current_run_meta[PYTHON_DEPENDENCY]
            pdep_list = dict_to_list(pdep)
            py_depen = {
                "data": pdep_list
            }

        # to get parameter and metric
        run_data = BaseParser.run_parser(BaseParser.get_parser_type(current_run_meta), current_run_meta)
        metric_data = run_data['metrics']
        parameter_data = run_data['params']
        visual_data = None
        if 'visual' in run_data:
            visual_data = run_data['visual']
        if current_run_meta[LcConst.RUN_OBJ_IS_EVALUATION]=="false":
            artifact = {
                "name": run_data['artifact']['name'],
                "version": run_data['artifact']['version']
            }

            artifact_proto = artifact
        else:
            artifact_proto = None

        req_body = json.dumps({
            "project_id": project_id,
            "workspace_id": workspace_id,
            "experiment_id": experiment_id,
            "userId": user_id,
            "run_nick_name": nick_name,
            "run": run_proto,  # Assuming this is now a dict
            "artifact": artifact_proto,  # Assuming this is now a dict
            "git": git_meta_data,
            "parameter": parameter_data,
            "metrics": metric_data,
            "visuals": visual_data,
            "dependency": py_depen
        })

        response = self._call_endpoint(req_body, current_run_meta)

        return response

    def create_tag(self, run_id, tag):
        """
        Set a tag for the specified run

        :param run_id: String ID of the run
        :param tag: RunTag instance to log
        """

        # TODO brian_todo

        pass

    def _make_metrics_data(self, json_body):
        result = []
        # metrics_object = {'key': '', 'values': [], 'timestamp': [], 'steps': []}

        # get metrics
        metrics_data = json_body['metrics']
        # get visuals
        if 'visuals' in json_body and json_body['visuals'] is not None:
            visuals_data = json_body['visuals']
            for item in visuals_data:
                metrics_object = {'key': item['key'], 'values': [], 'timestamp': item['timestamp'],
                                  'steps': item['steps']}
                # key, listValues, timestamp, steps
                if 'listValues' in item:
                    for values in item['listValues'][0]['values']:
                        metrics_object['values'].append(values['values'])
                else:
                    metrics_object['values'].extend(item['values'][0]['values'])

                result.append(metrics_object)
                metrics_data.append(metrics_object)

        if 'visuals' in json_body:
            del json_body['visuals']

        return json_body

    def _make_parameter_data(self, json_body):
        result = []
        # metrics_object = {'key': '', 'values': [], 'timestamp': [], 'steps': []}

        # get metrics
        try:
            parameter_data = json_body['parameter']
        except KeyError:
            parameter_data = []
        # "parameter": {
        #     "parameter": {
        #       "learningrate": "0.12",
        #       "subsample": "0.72",
        #       "colsamplebytree": "0.62",
        #       "maxdepth": "52",
        #       "alpha": "0.82"
        #     }
        #   }

        parameter = {'parameter': {}}

        parameter_dict = {k: v for d in parameter_data for k, v in d.items()}
        parameter = {'parameter': parameter_dict}
        #
        # for param in parameter_data:
        #     # key = param['key']
        #     # value = param['value']
        #     parameter['parameter'].append(param)
        #     # column_dict.setdefault(col_name, []).append(row[col_name])

        if 'parameter' in json_body:
            del json_body['parameter']

        json_body['parameter'] = parameter
        return json_body

    def _make_feature_data(self, json_body, run_info_json):
        json_body['feature'] = {}
        json_body['feature']['importance'] = {}

        run_result_path = run_info_json[RUN_RESULT_PATH]

        # json_path = run_result_path[RUN_MODEL_SHAP_JSON_PATH]

        # json_path = None

        if RUN_MODEL_SHAP_JSON_PATH in run_result_path.keys():
            json_path = run_result_path[RUN_MODEL_SHAP_JSON_PATH]

            with open(json_path) as json_file:
                shap_json_data = json.load(json_file)
            json_body['feature']['importance'] = shap_json_data

        return json_body


from Accuinsight.modeler.utils.dependency.dependencies import gather_sources_and_dependencies
from Accuinsight.modeler.core.LcConst import LcConst

def get_file_path(_caller_globals):
    print("Excuting environment", _caller_globals.get("get_ipython"))
    # print(_caller_globals)
    filepath = _caller_globals.get("__session__") or _caller_globals.get("__file__")
    print("call stack filename:", filepath)
    sources, dependencies = gather_sources_and_dependencies(globs=_caller_globals, save_git_info=False)

    if filepath.startswith(LcConst.ENV_JUPYTER_HOME_DIR+'/'):
        filepath_in_Jupyter = filepath[len(LcConst.ENV_JUPYTER_HOME_DIR+'/'):]
    return filepath_in_Jupyter, sources, dependencies


import json
import inspect

from Accuinsight.modeler.utils.os_getenv import get_os_env
from Accuinsight.modeler.core.get_for_visual import roc_pr_curve, get_true_y, get_visual_info_regressor
from Accuinsight.modeler.core import calculation_metrics
from Accuinsight.modeler.core.func import get_time
from Accuinsight.modeler.core.LcConst import LcConst
from Accuinsight.Lifecycle.utils import create_metrics_dict
from Accuinsight.modeler.clients.modeler_api_common import WorkspaceRestApi
from pathlib import Path
from collections import OrderedDict
from Accuinsight.modeler.utils.file_path import get_file_path


class RunData:
    def __init__(self, model_file_path, run_nickname_input, train_method, start_ts, end_ts):
        self.data = OrderedDict()
        self.data["path"] = model_file_path
        self.data["isEvaluation"] = "true"

        run_nick_name = Path(model_file_path).stem if run_nickname_input is None else run_nickname_input
        self.data["runNickName"] = run_nick_name
        self.data["name"] = run_nick_name + ' - ' + train_method + ' - No model information'

        self.data["userId"] = str(get_os_env()[LcConst.ENV_USER_SSO_ID])  # owner id
        self.data["creaDt"] = start_ts
        self.data["endDt"] = end_ts
        self.data["duration"] = end_ts - start_ts
        self.data["afLoc"] = ""
        self.data["note"] = ""
        self.data["status"] = "LcFINISHED"
        self.data["modelPath"] = ""
        self.data["jsonPath"] = ""

    def get(self):
        return self.data

class MetricsData:
    def __init__(self, train_method, y_hat_val, y_val, f1_average_in='weighted', recall_average_in='weighted'):
        self.data = []
        if train_method == 'classification':
            print("CLASSIFICATION in model_evaluation_from_data method")
            visual_classification_json = roc_pr_curve(y_hat_val, y_val, f1_average=f1_average_in, recall_average=recall_average_in)
            c_metrics = calculation_metrics.calculate_classification_metrics(y_hat_val, y_val, average=f1_average_in)

            self.data.extend([
                create_metrics_dict("fpr", list(value for value in visual_classification_json['fpr'].values()), list(visual_classification_json['fpr'].keys())),
                create_metrics_dict("tpr", list(value for value in visual_classification_json['tpr'].values()), list(visual_classification_json['tpr'].keys())),
                create_metrics_dict("recall", list(value for value in visual_classification_json['recall'].values()), list(visual_classification_json['recall'].keys())),
                create_metrics_dict("precision", list(value for value in visual_classification_json['precision'].values()), list(visual_classification_json['precision'].keys())),
                create_metrics_dict("legend",
                                    [[value['macro'], value['micro']] for value in visual_classification_json['legend'].values()],
                                    [f"{step}_{metric}" for step in visual_classification_json['legend'] for metric in ['macro', 'micro']]),
                create_metrics_dict("confusion_matrix", list(map(list, visual_classification_json['confusion_matrix'].values()))),
                create_metrics_dict("chart", list(visual_classification_json['chart'].values()), list(visual_classification_json['chart'].keys()))
            ])

        elif train_method == 'regression':
            print("REGRESSION in model_evaluation_from_data method")
            visual_regression_json = {'True_y': get_true_y(y_val), 'Predicted_y': get_visual_info_regressor(y_hat_val)}
            c_metrics = calculation_metrics.calculate_regression_metrics(y_hat_val, y_val)

            self.data.extend([
                create_metrics_dict("true_y", visual_regression_json['True_y']),
                create_metrics_dict("predicted_y", visual_regression_json['Predicted_y'])
            ])

        else:
            raise ValueError('현재 설정한 모델은 지원되지 않습니다. 관리자에게 문의하십시오.')

        selected_metrics_values = [str(metric_value) for metric_value in c_metrics.values()]
        selected_metrics = create_metrics_dict("selected_metrics", selected_metrics_values, steps=list(c_metrics.keys()))
        self.data.append(selected_metrics)

    def get(self):
        return self.data

class EnvironmentData:
    def __init__(self):
        self.data = OrderedDict()
        env_value = get_os_env()
        self.data["project_id"] = str(env_value[LcConst.ENV_PROJECT_ID])
        self.data["workspace_id"] = str(env_value[LcConst.ENV_WORKSPACE_ID])
        self.data["experiment_id"] = str(env_value[LcConst.ENV_EXPERIMENT_ID])
        self.data["userId"] = str(env_value[LcConst.ENV_USER_SSO_ID])  # owner id
        self.data["git"] = OrderedDict()
        self.data["git"]["url"] = ""
        self.data["git"]["commit"] = ""
        self.data["feature"] = OrderedDict()
        self.data["feature"]["importance"] = {}
        self.data["parameter"] = OrderedDict()
        self.data["parameter"]["parameter"] = {}
        self.data["dependency"] = OrderedDict()
        self.data["dependency"]["data"] = []

    def get(self):
        return self.data

def model_evaluation(run_nickname_input=None, train_method=None, y_hat_validation=None,
                     y_validation=None, f1_average_in='weighted', recall_average_in='weighted'):
    start_evl = get_time.now()

    _caller_globals = inspect.stack()[1][0].f_globals
    var_model_file_path, sources, dependencies = get_file_path(_caller_globals)

    if var_model_file_path is not None:
        print("filename: ", var_model_file_path)
    else:
        ValueError("filename을 찾을 수 없습니다.")

    run_data = RunData(var_model_file_path, run_nickname_input, train_method, int(start_evl.timestamp()),
                       int(get_time.now().timestamp()))
    metrics_data = MetricsData(train_method, y_hat_validation, y_validation, f1_average_in, recall_average_in)
    env_data = EnvironmentData()

    data = OrderedDict()
    data["run"] = run_data.get()
    data["metrics"] = metrics_data.get()
    data.update(env_data.get())

    env_value = get_os_env('ENV')
    modeler_rest = WorkspaceRestApi(env_value[LcConst.BACK_END_API_URL],
                                    env_value[LcConst.BACK_END_API_PORT],
                                    env_value[LcConst.BACK_END_API_URI])
    param = json.dumps(data)

    # modeler-api로 전송
    modeler_rest.call_rest_api(param, 'model_evaluation')


import inspect
import json
import math
from collections import OrderedDict
import warnings
import logging
from joblib import dump
from sklearn import pipeline

from Accuinsight.modeler.core import func, path, get
from Accuinsight.modeler.core.func import get_time
from Accuinsight.modeler.core.get_for_visual import roc_pr_curve, get_true_y, get_visual_info_regressor
from Accuinsight.modeler.core.sklearnModelType import REGRESSION, CLASSIFICATION
from Accuinsight.modeler.core.LcConst.LcConst import ALL_MODEL_PARAMS, SELECTED_PARAMS, SELECTED_METRICS, VALUE_ERROR, \
    LOGGING_TIME, LOGGING_RUN_ID, FITTED_MODEL, USER_ID, RUN_OBJ_NAME, RUN_MODEL_JSON_PATH, RUN_MODEL_VISUAL_JSON_PATH, \
    RUN_MODEL_SHAP_JSON_PATH, RUN_MODEL_JOBLIB_PATH, RUN_ID
from Accuinsight.modeler.core.Run.RunInfo.RunInfo import set_current_runs, clear_runs, \
    set_git_meta, set_python_dependencies, set_run_name, set_model_json_path, set_visual_json_path, set_model_file_path, \
    set_prefix_path, set_best_model_joblib_path, _set_result_path, set_shap_json_path, set_best_model_json_path
from Accuinsight.modeler.core.LcConst import LcConst
from Accuinsight.modeler.utils.os_getenv import get_current_notebook, get_os_env
from Accuinsight.modeler.clients.modeler_api import LifecycleRestApi

from Accuinsight.modeler.core.feature_contribution import shap_value
from Accuinsight.modeler.core import metricsLightgbm
from Accuinsight.Lifecycle.common import Common
from Accuinsight.modeler.utils.file_path import get_file_path


logging.basicConfig(level=logging.INFO,
                    format='%(message)s')

warnings.filterwarnings("ignore")


class accuinsight(Common):
    def __init__(self):
        global feature_name, alarm_api, message_param, user_id, \
            notebook_info, run_meta, run_nick_name, mail_alarm, mail_info
        super().__init__()

        feature_name = None
        user_id = None
        message_param = ''

        # alarm = self.workspace_alarm
        alarm_api = self.workspace_alarm_api
        notebook_info = self.notebook_info
        run_nick_name = self.default_run_nick_name
        mail_alarm = self.mail_alarm
        mail_info = self.mail_info

        run_meta = None

    def get_file(self, storage_json_file_name=None):
        global save_path, StorageInfo, target_name
        save_path, StorageInfo, target_name = super().get_file(storage_json_file_name)

    def set_features(self, feature_names):
        global feature_name
        feature_name = super().set_features(feature_names)

    def set_slack(self, hook_url=None):
        alarm_api.notifiers['slack'] = hook_url

    def unset_slack(self):
        if 'slack' in alarm_api.notifiers.keys():
            del alarm_api.notifiers['slack']

    # def set_mail(self, address=None):
    #     alarm.notifiers['mail'] = address
    #
    # def unset_mail(self):
    #     if 'mail' in alarm.notifiers.keys():
    #         del alarm.notifiers['mail']

    def unset_message(self):
        global message_param
        message_param = ''

    def set_user_id(self, uid):
        global user_id
        user_id = uid

    def unset_user_id(self):
        global user_id
        user_id = None

    def set_current_notebook(self):
        global notebook_info
        notebook_info = get_current_notebook()

    @staticmethod
    def get_current_run_meta():
        global run_meta
        return run_meta

    def get_current_run_id(self):
        try:
            return self.get_current_run_meta()[RUN_OBJ_NAME]
        except TypeError or KeyError:
            return None

    class add_experiment(object):
        def __init__(self, model_name, *args, model_monitor=False, runtime=False, f1_average_in='weighted', recall_average_in='weighted'):
            self.shap_on = model_monitor
            self.runtime = runtime
            self.input_nick_name = False
            logging.info('Using add_experiment(model_monitor={})'.format(model_monitor))
            if f1_average_in not in ['micro', 'macro', 'weighted']:
                raise ValueError("f1_average_in set micro|macro|weighted")
            if recall_average_in not in ['micro', 'macro', 'weighted']:
                raise ValueError("recall_average_in set micro|macro|weighted")

            if type(model_name) == pipeline.Pipeline:
                self.model_name = model_name.steps[-1][1]
            else:
                self.model_name = model_name

            _caller_globals = inspect.stack()[1][0].f_globals
            self.var_model_file_path, self.sources, self.dependencies = get_file_path(_caller_globals)

            if self.var_model_file_path is not None:
                print("filename: ", self.var_model_file_path)
            else:
                ValueError("filename을 찾을 수 없습니다.")

            self.fitted_model = get.model_type(self.model_name)          # fitted model type
            self.json_path = OrderedDict()                               # path
            self.selected_params = []                                    # log params
            self.selected_metrics = OrderedDict()                        # log metrics
            self.summary_info = OrderedDict()                            # final results
            self.error_log = []                                          # error log
            self.vis_info = None                                         # visualization info - classifier
            self.dict_path = path.get_file_path(self.model_name)

            set_current_runs(get.model_type(self.model_name))
            _set_result_path(self.dict_path)

            # data for visualization function
            if len(args) == 3:
                self.xval = args[1]
                self.yval = args[2]
                self.summary_info['run_nick_name'] = args[0]
                self.input_nick_name = True

            elif len(args) == 5:
                self.xtrain = args[1]
                self.ytrain = args[2]
                self.xval = args[3]
                self.yval = args[4]
                self.summary_info['run_nick_name'] = args[0]
                self.input_nick_name = True

            # 사용자가 run_nick_name 파라미터 입력 안했을 경우
            # ex1) with accu.add_experiment(model, X_train_scaled, y_train_num, X_test_scaled, y_test_num, model_monitor=True, runtime=True) as exp:
            # ex2) with accu.add_experiment(model, X_test_scaled, y_test_num, model_monitor=True, runtime=True) as exp:
            elif len(args) == 2:
                self.xval = args[0]
                self.yval = args[1]
                self.summary_info['run_nick_name'] = run_nick_name

            elif len(args) == 4:
                self.xtrain = args[0]
                self.ytrain = args[1]
                self.xval = args[2]
                self.yval = args[3]
                self.summary_info['run_nick_name'] = run_nick_name

            else:
                raise ValueError('Check the arguments of function - add_experiment(model_name, run_nick_name, X_val, y_val) or add_experiment(model_name, run_nick_name, X_train, y_train, X_val, y_val)',
                                 '\n and No assign run_nick_name, set default run_nick_name as File name.')

            # 사용자가 run_nick_name 파라미터를 입력했는데 '' 일 경우
            if self.input_nick_name == True and args[0] == '':
                self.summary_info['run_nick_name'] = run_nick_name

            # sklearn/xgboost/lightgbm
            get_from_model = get.from_model(self.model_name)
            self.all_model_params = get_from_model.all_params()

            if 'metric' not in self.all_model_params.keys():
                self.all_model_params['metric'] = ''

            self.model_param_keys = get_from_model.param_keys()
            
            # classifier
            if any(i in self.fitted_model for i in CLASSIFICATION):
                self.vis_info = roc_pr_curve(self.xval, self.yval, model=self.model_name ,f1_average=f1_average_in, recall_average=recall_average_in)

            # regressor
            elif any(i in self.fitted_model for i in REGRESSION):
                self.ypred = get_visual_info_regressor(self.xval, model=self.model_name)
            
            # if user uses lightgbm package, verify the model for classification or regression.
            if self.fitted_model == 'lightgbm':
                print('lightgbm.')
                # classifier
                if self.all_model_params['metric'] in metricsLightgbm.CLASSIFICATION:
                    self.vis_info = roc_pr_curve(self.xval, self.yval, model=self.model_name, f1_average=f1_average_in, recall_average=recall_average_in)
                
                # regressor
                elif self.all_model_params['metric'] in metricsLightgbm.REGRESSION:
                    self.ypred = get_visual_info_regressor(self.xval, model=self.model_name)
                
                # None
                elif self.all_model_params['metric'] in metricsLightgbm.no_metrics:
                    raise ValueError('Please set any metric parameter in the lightGBM.')
            
            set_model_file_path(self.var_model_file_path)

            if hasattr(self, 'mainfile'):
                set_git_meta(fileinfo=self.mainfile)
            if hasattr(self, 'dependencies'):
                set_python_dependencies(py_depenpency=self.dependencies)

        def __enter__(self):
            self.start_time = get_time.now()
            self.summary_info[LOGGING_TIME] = get_time.logging_time()
            self.summary_info[LOGGING_RUN_ID] = func.get_run_id()
            self.summary_info[FITTED_MODEL] = self.fitted_model

            if user_id is not None:
                self.summary_info[USER_ID] = user_id

            set_prefix_path(self.dict_path[LcConst.RUN_PREFIX_PATH])
            set_run_name(self.fitted_model, self.summary_info[LOGGING_RUN_ID], self.summary_info['run_nick_name'])

            return self

        def __exit__(self, a, b, c):
            self.summary_info[ALL_MODEL_PARAMS] = self.all_model_params
            for key in self.summary_info[ALL_MODEL_PARAMS]:
                if isinstance(self.summary_info[ALL_MODEL_PARAMS][key], float) and math.isnan(self.summary_info[ALL_MODEL_PARAMS][key]):
                    self.summary_info[ALL_MODEL_PARAMS][key] = None
            self.summary_info[SELECTED_PARAMS] = self.selected_params
            self.summary_info[SELECTED_METRICS] = self.selected_metrics
            self.summary_info[VALUE_ERROR] = self.error_log

            # model_monitor = True
            self.run_id = self.summary_info[LOGGING_RUN_ID]
            
            if self.shap_on:
                if not feature_name:
                    try:
                        if self.data is not None:
                            self.feature_name = get.feature_name(save_path, StorageInfo, target_name, data=self.data)
                            self.data = None
                        else:
                            self.feature_name = get.feature_name(save_path, StorageInfo, target_name)
                    except:
                        self.feature_name = None
                else:
                    self.feature_name = feature_name

                self.run_id = self.fitted_model + '-' + self.run_id
                
                # when user uses lightGBM fitted by lightGBM datasets, convert data type to lightGBM dataset.
                if self.fitted_model == 'lightgbm':
                    self.xtrain = lgb.Dataset(self.xtrain, label=self.ytrain, silent=True)
                        
                self.shap_value = shap_value(self.model_name, self.xtrain, self.feature_name)
                
                # path for shap.json
                shap_json_full_path = self.dict_path[RUN_MODEL_SHAP_JSON_PATH]
                set_shap_json_path(self.dict_path['shap_json'])
                
                with open(shap_json_full_path, 'w', encoding='utf-8') as save_file:
                    json.dump(self.shap_value, save_file, indent='\t')
                
            # visualization
            if any(i in self.fitted_model for i in CLASSIFICATION) or self.all_model_params['metric'] in metricsLightgbm.CLASSIFICATION:

                # path for visual.json
                visual_json_full_path = self.dict_path[RUN_MODEL_VISUAL_JSON_PATH]
                set_visual_json_path(self.dict_path['visual_json'])

                with open(visual_json_full_path, 'w', encoding='utf-8') as save_file1:
                    json.dump(self.vis_info, save_file1, indent="\t")

            elif any(i in self.fitted_model for i in REGRESSION) or self.all_model_params['metric'] in metricsLightgbm.REGRESSION:
                temp_yval = get_true_y(self.yval)
                if len(temp_yval) <= 5000:
                    self.summary_info['True_y'] = temp_yval
                    self.summary_info['Predicted_y'] = get_visual_info_regressor(self.xval, model=self.model_name)
                else:
                    self.summary_info['True_y'] = None
                    self.summary_info['Predicted_y'] = None

            self.summary_info[VALUE_ERROR] = self.error_log

            if not self.summary_info[VALUE_ERROR]:
                # path for model_info.json
                model_json_full_path = self.dict_path[RUN_MODEL_JSON_PATH]
                set_model_json_path(self.dict_path['model_json'])

                with open(model_json_full_path, 'w', encoding='utf-8') as save_file2:
                    json.dump(self.summary_info, save_file2, indent="\t")
                    
            else:
                pass

            # model save
            save_model_path = self.dict_path[RUN_MODEL_JOBLIB_PATH] + self.summary_info[FITTED_MODEL] + '-' + self.summary_info[LOGGING_RUN_ID] +'.joblib'
            path_for_setting_model_json = self.dict_path['save_model_dir'] + '/' + self.summary_info[FITTED_MODEL] + '-' + self.summary_info[LOGGING_RUN_ID] +'.json'
            path_for_setting_model_joblib = self.dict_path['save_model_dir'] + '/' + self.summary_info[FITTED_MODEL] + '-' + self.summary_info[LOGGING_RUN_ID] +'.joblib'

            # /best_model/ directory 저장하고 나중에 불러와서 run_info에 set할때 run/ 붙임.
            browser_common_path = self.dict_path['save_model_dir'] + '/' + self.summary_info[FITTED_MODEL] + '-' + self.summary_info[LOGGING_RUN_ID]
            browser_model_json_path = browser_common_path + '.json'
            browser_model_joblib_path = browser_common_path + '.joblib'

            # best model save as JSON
            summary_info_json_for_evl = self.summary_info.copy()
            summary_info_json_for_evl[ALL_MODEL_PARAMS] = self.summary_info[ALL_MODEL_PARAMS]
            summary_info_json_for_evl[SELECTED_PARAMS] = self.summary_info[SELECTED_PARAMS]
            summary_info_json_for_evl[RUN_ID] = self.summary_info[LOGGING_RUN_ID]
            summary_info_json_for_evl['path'] = self.var_model_file_path
            summary_info_json_for_evl[FITTED_MODEL] = self.summary_info[FITTED_MODEL]
            summary_info_json_for_evl['model_dependencies'] = self.dependencies
            summary_info_json_for_evl['best_model_joblib'] = browser_model_joblib_path
            summary_info_json_for_evl['best_model_json'] = browser_model_json_path
            save_model_json_path = self.dict_path[RUN_MODEL_JOBLIB_PATH] + self.summary_info[FITTED_MODEL] + '-' + self.summary_info[LOGGING_RUN_ID] +'.json'
            with open(save_model_json_path, 'w', encoding='utf-8') as json_file:
                json.dump(summary_info_json_for_evl, json_file, indent="\t")

            set_best_model_json_path(path_for_setting_model_json)
            set_best_model_joblib_path(path_for_setting_model_joblib)

            dump(self.model_name, save_model_path)

            start_time = int(self.start_time.timestamp()*1000)
            end_time = int(get_time.now().timestamp()*1000)
            delta_ts = end_time - start_time

            global run_meta
            run_meta = clear_runs(start_time, end_time, delta_ts)

            accuinsight._send_message(metric=None,
                                      current_value=None,
                                      message=message_param,
                                      thresholds=None,
                                      alarm_object=None, #modeler alarm api deprecated
                                      alarm_api=alarm_api)

            env_value = get_os_env('ENV')
            modeler_rest = LifecycleRestApi(env_value[LcConst.BACK_END_API_URL],
                                            env_value[LcConst.BACK_END_API_PORT],
                                            env_value[LcConst.BACK_END_API_URI])
            modeler_rest.lc_create_run(run_meta)
            if self.runtime:
                accuinsight.set_runtime_model('sklearn')

        def log_params(self, param=None):
            # sklearn/xgboost/lightgbm
            if param:
                if param in self.model_param_keys:
                    return self.selected_params.append(param)

                else:
                    self.error_log.append(True)
                    raise ValueError('"' + param + '"' + ' does not exist in the model.')

        def log_metrics(self, metric_name, defined_metric):
            self.selected_metrics[metric_name] = defined_metric

        def log_tag(self, description):
            self.summary_info['tag'] = description



import json
from Accuinsight.modeler.core.LcConst.LcConst import RUN_RESULT_PATH, RUN_MODEL_JSON_PATH
from Accuinsight.modeler.core.LcConst.LcConst import ALL_MODEL_PARAMS, SELECTED_METRICS, SELECTED_PARAMS
from Accuinsight.modeler.protos.life_cycle_pb2 import LcParam, LcMetric
from Accuinsight.modeler.core.Run.ParseRun import ParserVisualJaon, Parse_Helper

def _makeLcParam(selected_params, all_model_params):
    grpc_params = []

    for param_key in selected_params:

        param = LcParam(
            key=param_key,
            value=str(all_model_params[param_key])
        )
        grpc_params.append(param)
    return grpc_params


# parse parameters
def parse_run_parameters(json_data):
    # read ~/runs/results-XGBClassifier/model-info-json/*.json
    # return parameters

    param_data = json_data
    selected_params = param_data[SELECTED_PARAMS]
    all_model_params = param_data[ALL_MODEL_PARAMS]

    grpc_params = _makeLcParam(selected_params=selected_params, all_model_params=all_model_params)

    return grpc_params


# parse metrics
def parse_run_metrics(json_data):
    # read ~/runs/results-XGBClassifier/model-info-json/*.json
    # get "selected_metrics" field
    # return metrics

    metric_data = json_data
    grpc_metrics = []

    # selected_metrics
    selected_metrics = metric_data[SELECTED_METRICS]
    metric_keys = selected_metrics.keys()

    result_dict = {'metrics': []}

    count = 0
    timestamp = []
    steps = []
    values = []
    for key in metric_keys:
        values.append(str(selected_metrics[key]))
        timestamp.append(str(count))
        steps.append(key)
        count += 1

    result_dict['metrics'].append({
        'values': values,
        'timestamp': timestamp,
        'steps': steps})

    metrics = result_dict['metrics']

    for item in metrics:
        metric = LcMetric(
            key=SELECTED_METRICS,
            values=item['values'],
            timestamp=item['timestamp'],
            steps=item['steps']
        )
        grpc_metrics.append(metric)

    return grpc_metrics

def parse_run_result(run_info_json):
    result_dict = {'metrics': '', 'params': ''}

    grpc_params = parse_run_parameters(json_data=run_info_json)
    grpc_metrics = parse_run_metrics(json_data=run_info_json)

    grpc_visuals = ParserVisualJaon.parse_run_visual(run_mata=run_info_json)
    result_dict['params'] = grpc_params
    result_dict['metrics'] = grpc_metrics
    result_dict['visual'] = grpc_visuals

    return result_dict


"""
This http client send run's results data to modeler back-end server
"""

from Accuinsight.modeler.store.tracking.life_cycle_store import RestStore as LcRestStore
from Accuinsight.modeler.utils.os_getenv import get_os_env
from Accuinsight.modeler.core.LcConst import LcConst

# api for experiment
class LifecycleRestApi:
    def __init__(self, host_url, port, uri):
        self.base_url = 'http://' + host_url + ':' + str(port) + '/' + uri

    def lc_create_run(self, current_run_meta):
        """Create run."""

        store = LcRestStore(self.ModelerHostCreds(self.base_url))
        run = store.lc_create_run(current_run_meta)
        return run
    class ModelerHostCreds(object):
        """
        Provides a hostname and optional authentication for talking to an back-end server.
        :param host: Hostname (e.g., http://localhost:5000) to back-end server. Required.
        :param username: Username to use with Basic authentication when talking to server.
            If this is specified, password must also be specified.
        :param password: Password to use with Basic authentication when talking to server.
            If this is specified, username must also be specified.
        :param token: Token to use with Bearer authentication when talking to server.
            If provided, user/password authentication will be ignored.
        :param ignore_tls_verification: If true, we will not verify the server's hostname or TLS
            certificate. This is useful for certain testing situations, but should never be
            true in production.
        """

        def __init__(self, host, username=None, password=None, token=None,
                     ignore_tls_verification=False):
            if not host:
                raise ValueError("host is a required parameter for ModelerHostCreds")
            self.host = host
            self.username = username
            self.password = password
            self.token = token
            self.ignore_tls_verification = ignore_tls_verification


# api for model deployment (deprecated on 3.0)
# class DeployLogRestApi:
#     def __init__(self, host_url, port, uri):
#         self.base_url = 'http://' + host_url + ':' + str(port) + '/' + uri
#
#     def call_rest_api(self, method, param, mode):
#         store = MonitoringDeployRestStore(ModelerHostCreds(self.base_url))
#         response = store.call_endpoint(method, param, mode)
#
#         return response


if __name__ == "__main__":
    env_value = get_os_env('ENV')
    modeler_rest = LifecycleRestApi(env_value[LcConst.BACK_END_API_URL],
                                    env_value[LcConst.BACK_END_API_PORT],
                                    env_value[LcConst.BACK_END_API_URI])
    modeler_rest.lc_create_run()



import json
from Accuinsight.modeler.core.LcConst.LcConst import RUN_BASE_PATH, RUN_PREFIX_PATH, ALL_MODEL_PARAMS, \
    SELECTED_METRICS, SELECTED_PARAMS, RUN_OBJ_MODEL_JSON_PATH
from Accuinsight.modeler.protos.life_cycle_pb2 import LcParam, LcMetric
from Accuinsight.modeler.core.Run.ParseRun import Parse_Helper

def _makeLcParam(selected_params, all_model_params):
    grpc_params = []

    for param_key in selected_params:

        param = LcParam(
            key=param_key,
            value=str(all_model_params[param_key])
        )
        grpc_params.append(param)
    return grpc_params


def _append_metrics(metrics, **kwargs):
    values = []
    timestamp = []
    steps = []
    count = 0
    data = kwargs['data']
    key = kwargs['key']

    for v in data:
        values.append(str(v))
        timestamp.append(str(count))
        steps.append(str(count))
        count += 1
    metrics.append({'key': key,
                    'values': values,
                    'timestamp': timestamp,
                    'steps': steps})
    return metrics


def _parse_true_predict(metric_data, metrics):
    true_y = []
    predicted_y = []
    true_y_values = metric_data['True_y']
    predicted_y_values = metric_data['Predicted_y']

    # check if nested list
    is_nested_true = any(isinstance(elem, list) for elem in true_y_values)
    is_nested_predicted = any(isinstance(elem, list) for elem in predicted_y_values)

    if is_nested_true:
        for elem in true_y_values:
            true_y.append(elem[0])
    else:
        true_y = true_y_values

    if is_nested_predicted:
        for elem in predicted_y_values:
            predicted_y.append(elem[0])
    else:
        predicted_y = predicted_y_values

    metrics = _append_metrics(metrics, data=true_y, key='true_y')
    metrics = _append_metrics(metrics, data=predicted_y, key='predicted_y')
    return metrics


# parse parameters
def parse_run_parameters(json_data):
    # read ~/runs/results-XGBClassifier/model-info-json/*.json
    # return parameters

    param_data = json_data
    selected_params = param_data[SELECTED_PARAMS]
    all_model_params = param_data[ALL_MODEL_PARAMS]

    grpc_params = _makeLcParam(selected_params=selected_params, all_model_params=all_model_params)

    return grpc_params


# parse metrics
def _parse_selected_metrics(json_data):
    # read ~/runs/results-XGBClassifier/model-info-json/*.json
    # get "selected_metrics" field
    # return metrics

    metric_data = json_data

    grpc_metrics = []
    if SELECTED_METRICS not in metric_data:
        return grpc_metrics

    # selected_metrics
    selected_metrics = metric_data[SELECTED_METRICS]
    metric_keys = selected_metrics.keys()

    result_dict = {'metrics': []}

    count = 0
    timestamp = []
    steps = []
    values = []
    for key in metric_keys:
        values.append(str(selected_metrics[key]))
        timestamp.append(str(count))
        steps.append(key)
        count += 1

    result_dict['metrics'].append({
                                   'values': values,
                                   'timestamp': timestamp,
                                   'steps': steps})

    metrics = result_dict['metrics']

    for item in metrics:
        metric = LcMetric(
                key=SELECTED_METRICS,
                values=item['values'],
                timestamp=item['timestamp'],
                steps=item['steps']
            )
        grpc_metrics.append(metric)

    return grpc_metrics


def parse_run_metrics(json_data):
    # read ~/runs/results-XGBClassifier/model-info-json/*.json
    # get "selected_metrics" field
    # return metrics

    metrics = []
    metrics = _parse_true_predict(json_data, metrics)

    grpc_metrics = []

    for item in metrics:
        metric = LcMetric(
                key=item['key'],
                values=item['values'],
                timestamp=item['timestamp'],
                steps=item['steps']
            )
        grpc_metrics.append(metric)

    selected_metrics = _parse_selected_metrics(json_data=json_data)
    grpc_metrics = grpc_metrics + selected_metrics

    return grpc_metrics


# parse metrics
def _append_visual_data(visual_data, key):
    result_dict = {'metrics': []}

    count = 0
    timestamp = []
    steps = []
    values = []

    for k, v in visual_data.items():
        timestamp.append(str(count))
        count += 1

        steps.append(k)
        values.append(v)

    result_dict['metrics'].append({'key': key,
                                   'values': values,
                                   'timestamp': timestamp,
                                   'steps': steps})
    return result_dict


def _make_steps(steps_value):
    steps_list = []
    for value in steps_value:
        steps_list.append(str(value))
    return steps_list


def parse_run_result(run_info_json):
    result_dict = {'metrics': '', 'params': ''}

    grpc_params = parse_run_parameters(json_data=run_info_json)
    grpc_metrics = parse_run_metrics(json_data=run_info_json)

    result_dict['params'] = grpc_params
    result_dict['metrics'] = grpc_metrics

    return result_dict


import os
import shutil
import sys
import json
import logging
import time
import datetime
from collections import OrderedDict
from pathlib import Path
import requests

import boto3
import joblib

from Accuinsight.Lifecycle.utils import load_model
from Accuinsight.modeler.core.LcConst import LcConst
from Accuinsight.modeler.core.LcConst.LcConst import ENV_JUPYTER_HOME_DIR
from Accuinsight.modeler.utils.runs_utils import ProgressPercentage
from Accuinsight.modeler.clients.modeler_api_common import WorkspaceRestApi
from Accuinsight.modeler.utils.os_getenv import get_os_env, get_current_notebook
# from Accuinsight.modeler.entities.alarm import Alarm

class Common(object):
    def __init__(self):
        env_value = get_os_env('ENV')

        self.StorageInfo = None
        self.target_name = None
        self.save_path = None
        self.endpoint = None
        self.thresholds = None
        self.hook_url = None
        self.message = None
        self.feature_name = None
        self.data = None
        self.user_id = None
        self.storage_info_from_set = OrderedDict()
        self.workspace_alarm_api = WorkspaceRestApi(env_value[LcConst.BACK_END_API_URL],
                                                    env_value[LcConst.BACK_END_API_PORT],
                                                    env_value[LcConst.BACK_END_API_URI])
        # self.workspace_alarm = Alarm()
        self.mail_alarm = False
        self.mail_info = None

        self.notebook_info = get_current_notebook()
        self.default_run_nick_name = Path(self.notebook_info).stem
    def set_storage(self,
                    access_key=None,  # aws-s3
                    secret_key=None,  # aws-s3
                    region=None,  # aws-s3
                    bucket_name=None,  # aws-s3
                    endpoint_url=None,  # none aws-s3
                    file_path=None,  # common
                    target=None,  # common
                    save_json=False,  # common - if it is TRUE, saving a storage information as a json file.
                    save_path=None  # common - if save_path is not None, the json file is saved in this path.
                    ):

        if target is None:
            raise ValueError('target을 입력해주시기 바랍니다.')

        ## make a directory for saving json file
        save_prefix = None
        if save_json:
            save_prefix_run = os.path.join(ENV_JUPYTER_HOME_DIR, 'runs')
            if not os.path.isdir(save_prefix_run):
                try:
                    os.mkdir(save_prefix_run)
                except:
                    pass
            save_prefix = os.path.join(save_prefix_run, 'storage-info-json')
            if not os.path.isdir(save_prefix):
                try:
                    os.mkdir(save_prefix)
                except:
                    pass

        ### S3
        elif bucket_name is not None:
            ### AWS - S3
            if endpoint_url is None:
                storage_type = 's3'
                self.storage_info_from_set['myAccessKey'] = access_key
                self.storage_info_from_set['mySecretKey'] = secret_key
                self.storage_info_from_set['endpoint'] = 's3.' + region + '.amazonaws.com'
                self.storage_info_from_set['region'] = region
                self.storage_info_from_set['bucketType'] = 's3'
                self.storage_info_from_set['bucketName'] = bucket_name
                self.storage_info_from_set['filePath'] = file_path
                self.storage_info_from_set['fileName'] = file_path.split('/')[-1]
                self.storage_info_from_set['target'] = target

            ### other - S3
            else:
                storage_type = 's3'
                self.storage_info_from_set['myAccessKey'] = access_key
                self.storage_info_from_set['mySecretKey'] = secret_key
                self.storage_info_from_set['endpointUrl'] = endpoint_url
                self.storage_info_from_set['region'] = region
                self.storage_info_from_set['bucketType'] = 's3'
                self.storage_info_from_set['bucketName'] = bucket_name
                self.storage_info_from_set['filePath'] = file_path
                self.storage_info_from_set['fileName'] = file_path.split('/')[-1]
                self.storage_info_from_set['target'] = target

        ### Local
        else:
            storage_type = 'local'
            self.storage_info_from_set['filePath'] = file_path
            self.storage_info_from_set['fileName'] = file_path.split('/')[-1]
            self.storage_info_from_set['target'] = target

            self.save_path = file_path
            self.StorageInfo = self.storage_info_from_set
            self.target_name = target

        # save_json 만 True로 하고 사용자가 파라미터로 save_path를 입력하지 않을 경우
        if save_json:
            if save_path is None:
                timestamp = str(datetime.datetime.now().date()).replace('-', '')
                dir_list = os.listdir(save_prefix)
                num = len([True for i in dir_list if timestamp in i])
                save_path = timestamp + '-' + str(num + 1).zfill(2) + '-' + storage_type + '-connection-info' + '.json'

            else:  # save_path is not None
                pass

            storage_info_full_path = os.path.join(save_prefix, save_path)
            with open(storage_info_full_path, 'w', encoding='utf-8') as save_file:
                json.dump(self.storage_info_from_set, save_file, indent='\t')

        if not save_json and save_path is not None:
            raise ValueError("'save_json=True'로 설정하시기 바랍니다.")

    def get_file(self, storage_json_file_name=None):
        if storage_json_file_name is not None:

            # meta file 저장하는 폴더 만들면, 이 경로는 수정돼야 함.
            storage_info_json_path = os.path.join(ENV_JUPYTER_HOME_DIR, storage_json_file_name)

            with open(storage_info_json_path) as jsonFile:
                self.StorageInfo = json.load(jsonFile)

        else:  # storage_json_file_name == None, when using set_storage() method.
            self.StorageInfo = self.storage_info_from_set

        ## define the target name
        self.target_name = self.StorageInfo['target']

        ### HDFS 사용 안함
        # if 'host' in self.StorageInfo.keys():

        #     ## path for saving data
        #     save_dir = os.path.join(ENV_JUPYTER_HOME_DIR, 'data_from_hdfs')
        #     if not os.path.exists(save_dir):
        #         os.mkdir(save_dir)
        #     else:
        #         pass

        #     split_list = list(filter(None, self.StorageInfo['filePath'].split('/')))

        #     HOST = self.StorageInfo['host']
        #     PORT = int(self.StorageInfo['port'])
        #     FILE_PATH = self.StorageInfo['filePath']
        #     FILE_DIR = '/'
        #     for i in split_list[:len(split_list) - 1]:
        #         FILE_DIR = os.path.join(FILE_DIR, i)
        #     FILE_NAME = split_list[-1]

        #     save_file_name = FILE_NAME.split('.')[0] + '_' + str(datetime.datetime.now().date()).replace('-', '') + '.'\
        #                      + FILE_NAME.split('.')[1]
        #     self.save_path = os.path.join(save_dir, save_file_name)
        #     self.endpoint = os.path.join('hdfs://', HOST, self.StorageInfo['port']) + FILE_PATH

        #     hdfs = HDFileSystem(host=HOST, port=PORT)
        #     hdfs.ls(FILE_DIR)

        #     sys.stdout.write('%s %s %s' % ('Downloading file...', FILE_NAME, '\n'))
        #     time.sleep(1)
        #     hdfs.get(FILE_PATH, self.save_path)
        #     logging.info(self.save_path)

        ### AWS-S3
        if 'bucketName' in self.StorageInfo.keys():
            if not 'endpointUrl' in self.StorageInfo.keys():
                BUCKET_TYPE = self.StorageInfo['bucketType']
                BUCKET_NAME = self.StorageInfo['bucketName']
                FILE_PATH = self.StorageInfo['filePath']
                FILE_NAME = self.StorageInfo['fileName']
                # FILE_TYPE = self.StorageInfo['fileType']
                # FILE_DELIM = self.StorageInfo['fileDelim']
                ACCESS_KEY = self.StorageInfo['myAccessKey']
                SECRET_KEY = self.StorageInfo['mySecretKey']
                REGION = self.StorageInfo['region']
                URL = self.StorageInfo['endpoint']
                ## path for saving data
                save_dir = os.path.join(ENV_JUPYTER_HOME_DIR, 'data_from_aws')
                if not os.path.exists(save_dir):
                    os.mkdir(save_dir)
                else:
                    pass

                save_file_name = FILE_NAME.split('.')[0] + '_' + str(datetime.datetime.now().date()).replace('-', '') \
                                 + '.' + FILE_NAME.split('.')[1]
                self.save_path = os.path.join(save_dir, save_file_name)

                # endpoint
                pre_url = 'https://' + BUCKET_NAME + '.' + URL
                self.endpoint = os.path.join(pre_url, FILE_PATH)

                client = boto3.client(BUCKET_TYPE,
                                      aws_access_key_id=ACCESS_KEY,
                                      aws_secret_access_key=SECRET_KEY,
                                      region_name=REGION)

                transfer = boto3.s3.transfer.S3Transfer(client)

                progress = ProgressPercentage(client, BUCKET_NAME, FILE_PATH)

                sys.stdout.write('%s %s %s' % ('Downloading file...', FILE_NAME, '\n'))
                transfer.download_file(BUCKET_NAME, FILE_PATH, self.save_path, callback=progress)
                logging.info(self.save_path)

            else:
                BUCKET_TYPE = self.StorageInfo['bucketType']
                BUCKET_NAME = self.StorageInfo['bucketName']
                FILE_PATH = self.StorageInfo['filePath']
                FILE_NAME = self.StorageInfo['fileName']
                ACCESS_KEY = self.StorageInfo['myAccessKey']
                SECRET_KEY = self.StorageInfo['mySecretKey']
                REGION = self.StorageInfo['region']
                ENDPOINT_URL = self.StorageInfo['endpointUrl']
                ## path for saving data
                save_dir = os.path.join(ENV_JUPYTER_HOME_DIR, 'data_from_other')
                if not os.path.exists(save_dir):
                    os.mkdir(save_dir)
                else:
                    pass

                save_file_name = FILE_NAME.split('.')[0] + '_' + str(datetime.datetime.now().date()).replace('-', '')\
                                      + '.' + FILE_NAME.split('.')[1]
                self.save_path = os.path.join(save_dir, save_file_name)

                # endpoint
                client = boto3.client(BUCKET_TYPE,
                                      aws_access_key_id=ACCESS_KEY,
                                      aws_secret_access_key=SECRET_KEY,
                                      region_name=REGION,
                                      endpoint_url=ENDPOINT_URL)

                transfer = boto3.s3.transfer.S3Transfer(client)

                progress = ProgressPercentage(client, BUCKET_NAME, FILE_PATH)

                sys.stdout.write('%s %s %s' % ('Downloading file...', FILE_NAME, '\n'))
                transfer.download_file(BUCKET_NAME, FILE_PATH, self.save_path, callback=progress)
                logging.info(self.save_path)

        ### Local
        else:
            self.save_path = self.StorageInfo['filePath']

        return self.save_path, self.StorageInfo, self.target_name

    def set_data(self, data=None):
        if data is not None:
            self.data = data

    def unset_data(self):
        self.data = None

    @staticmethod
    def _send_message(metric=None, current_value=None, message=None, thresholds=None, alarm_object=None, alarm_api=None):
        if alarm_object:
            messages = list()

            if thresholds:
                if current_value >= thresholds:
                    messages.append('[모델 학습 완료] ' + metric + '이 설정하신 thresolds: ' + str(thresholds) + '를 초과하였습니다.')

            if message:
                if isinstance(message, str):
                    messages.append(message)
                elif isinstance(message, list):
                    messages.extend(message)
                else:
                    raise ValueError("Wrong message input detected. Available type is 'str' and 'list'")

            if messages:
                for message in messages:
                    alarm_object.message = message
                    alarm_object.type = "Workspace"
                    alarm_api.call_rest_api(alarm_object.get_alarm_param(), "alarm")

    def set_slack(self, hook_url=None):
        if hook_url:
            self.workspace_alarm.notifiers['slack'] = hook_url

    def unset_slack(self):
        if 'slack' in self.workspace_alarm.notifiers.keys():
            del self.workspace_alarm.notifiers['slack']

    # 기존 modeler-api를 통한 메일 기능 주석처리
    # def set_mail(self, address=None):
    #     if address:
    #         self.workspace_alarm.notifiers['mail'] = address
    #
    # def unset_mail(self):
    #     if 'mail' in self.workspace_alarm.notifiers.keys():
    #         del self.workspace_alarm.notifiers['mail']

    ''' setting example
    accu.set_email(to='aaa@sk.com, bbb@sk.com', cc='aaa@sk.com, bbb@sk.com', bcc='aaa@sk.com, bbb@sk.com',
                   subject='test mail from accu sdk', contents='something')
    '''

    @staticmethod
    def send_mail(mail_url='http://accumail-master.aiip-common.svc.cluster.local:8000/api/send_email/', to=None, cc=None, bcc=None, subject=None, message=None):
        if mail_url is None:
            raise ValueError("mail_url을 설정해야 합니다.")
        if to is None and cc is None and bcc is None:
            raise ValueError("메일 수신 설정을 확인하세요.")
        if subject is None:
            raise ValueError("메일 제목을 입력해주세요.")
        if message is None:
            raise ValueError("메일 내용을 입력해주세요.")
        headers = {
            "Content-Type": "application/json"
        }
        data = {
            "to": to,
            "cc": cc,
            "bcc": bcc,
            "subject": subject,
            "contents": f"{message}"
        }

        response = requests.post(mail_url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            print("메일이 성공적으로 전송되었습니다.")
        else:
            print("메일 전송 중 오류가 발생했습니다.")

    def send_message(self, message=None, thresholds=None):
        if message and thresholds:
            raise ValueError("'message'와 'thresholds' 두 개의 arguments를 동시에 입력할 수 없습니다.")
        else:
            self.message = message
            self.thresholds = thresholds

    def unset_message(self):
        self.message = None
        self.thresholds = None

    def set_user_id(self, user_id):
        self.user_id = user_id

    def unset_user_id(self):
        self.user_id = None

    def set_features(self, feature_names):
        if str(type(feature_names)) == "<class 'list'>":
            self.feature_name = feature_names
        else:
            try:
                self.feature_name = feature_names.tolist()
            except:
                pass

            try:
                self.feature_name = list(feature_names)
            except AttributeError:
                raise ValueError('입력값으로 list 또는 pandas.core.indexes.base.Index를 사용하시기 바랍니다.')

        return feature_names

    def set_current_notebook(self):
        self.notebook_info = get_current_notebook()

    def get_current_notebook(self):
        return self.notebook_info

    @staticmethod
    def set_runtime_model(framework, run_id=None):
        if framework == 'sklearn':
            if run_id:
                export_path = ENV_JUPYTER_HOME_DIR + '/airuntime/' + framework + '/models'
                shutil.rmtree(export_path)
                os.mkdir(export_path)
                joblib.dump(load_model(run_id), export_path + '/model.joblib')
            else:
                with open(ENV_JUPYTER_HOME_DIR + '/runs/run_info.json') as f:
                    json_object = json.load(f)

                try:
                    shutil.copyfile(ENV_JUPYTER_HOME_DIR + '/' + json_object['save_best_model_joblib_path'],
                                    ENV_JUPYTER_HOME_DIR + '/airuntime/' + framework + '/models/model.joblib')
                except TypeError:
                    raise ValueError('최근에 실행된 experiment가 ML 모델이 아닙니다.')

        elif framework == 'tensorflow' or framework == 'keras':
            if run_id:
                if framework == 'keras':
                    framework = 'tensorflow'

                saved_model = load_model(run_id)

            else:
                with open(ENV_JUPYTER_HOME_DIR + '/runs/run_info.json') as f:
                    json_object = json.load(f)

                try:
                    with open(ENV_JUPYTER_HOME_DIR + '/' + json_object['save_best_model_json_path'], 'r') as f:
                        saved_model_json = f.read()
                except TypeError:
                    raise ValueError('최근에 실행된 experiment가 DL 모델이 아닙니다.')

                if framework == 'keras':
                    from keras.models import model_from_json
                    saved_model = model_from_json(saved_model_json)
                    framework = 'tensorflow'
                else:
                    from tensorflow.keras.models import model_from_json
                    saved_model = model_from_json(saved_model_json)

                # load weights into new model
                saved_model.load_weights(ENV_JUPYTER_HOME_DIR + '/' + json_object['save_best_model_h5_path'])

            export_path = ENV_JUPYTER_HOME_DIR + '/airuntime/' + framework + '/models'
            shutil.rmtree(export_path) # 가장 최근에 복사된 모델 삭제
            saved_model.save(export_path + '/1', save_format='tf')
        else:
            raise ValueError("잘못된 framework 값이 입력되었습니다. framework가 가질 수 있는 값은 sklearn/tensorflow/keras 입니다.")



import json
from enum import Enum
from Accuinsight.modeler.core.LcConst import LcConst
from Accuinsight.modeler.core.Run.ParseRun import Parse_MLClassifier
from Accuinsight.modeler.core.Run.ParseRun import Parse_DLRegression
from Accuinsight.modeler.core.Run.ParseRun import Parse_DLClassification
from Accuinsight.modeler.core.Run.ParseRun import Parse_MLRegression

# from Accuinsight.modeler.core.Run.ParseRun import Parse_MLClassifier_evl
# from Accuinsight.modeler.core.Run.ParseRun import Parse_MLRegression_evl
# from Accuinsight.modeler.core.Run.ParseRun import Parse_DLClassification_evl
# from Accuinsight.modeler.core.Run.ParseRun import Parse_DLRegression_evl


class ModelType(Enum):
    DL_REGRESSION = 1
    DL_CLASSIFICATION = 2
    ML_REGRESSION = 3
    ML_CLASSIFICATION = 4
    DL_REGRESSION_evl = 5
    DL_CLASSIFICATION_evl = 6
    ML_REGRESSION_evl = 7
    ML_CLASSIFICATION_evl = 8


def run_parser(parser_type, run_info_json):
    if parser_type is ModelType.DL_REGRESSION:
        return Parse_DLRegression.parse_run_result(run_info_json)

    if parser_type is ModelType.DL_CLASSIFICATION:
        return Parse_DLClassification.parse_run_result(run_info_json)

    if parser_type is ModelType.ML_REGRESSION:
        return Parse_MLRegression.parse_run_result(run_info_json)

    if parser_type is ModelType.ML_CLASSIFICATION:
        return Parse_MLClassifier.parse_run_result(run_info_json)

    #evaluation
    # if parser_type is ModelType.DL_CLASSIFICATION_evl:
    #     return Parse_DLClassification_evl.parse_run_result(run_info_json)
    #
    # if parser_type is ModelType.DL_REGRESSION_evl:
    #     return Parse_DLRegression_evl.parse_run_result(run_info_json)
    #
    # if parser_type is ModelType.ML_CLASSIFICATION_evl:
    #     return Parse_MLClassifier_evl.parse_run_result(run_info_json)
    #
    # if parser_type is ModelType.ML_REGRESSION_evl:
    #     return Parse_MLRegression_evl.parse_run_result(run_info_json)
    else:
        print('Data parsing error for rest API (lc_create_run method)')
        exit(1)

def get_parser_type(run_info_json):
    parser_type = None
    run_result_path = run_info_json[LcConst.RUN_RESULT_PATH]

    # data_path = run_result_path[LcConst.RUN_MODEL_JSON_PATH]
    visual_json_path = None
    visual_csv_path = None
    evl_json_path = None

    if LcConst.RUN_MODEL_VISUAL_JSON_PATH in run_result_path:
        visual_json_path = run_result_path[LcConst.RUN_MODEL_VISUAL_JSON_PATH]

    if LcConst.RUN_MODEL_VISUAL_CSV_PATH in run_result_path:
        visual_csv_path = run_result_path[LcConst.RUN_MODEL_VISUAL_CSV_PATH]

    # determine parser type
    if visual_json_path is None and visual_csv_path is not None:
        parser_type = ModelType.DL_REGRESSION

    if visual_json_path is not None and visual_csv_path is not None:
        parser_type = ModelType.DL_CLASSIFICATION

        visual_json_path = run_result_path[LcConst.RUN_MODEL_VISUAL_JSON_PATH]
        with open(visual_json_path) as json_file:
            visual_json_data = json.load(json_file)

        if LcConst.TRUE_Y in visual_json_data:
            parser_type = ModelType.DL_REGRESSION

    if visual_json_path is None and visual_csv_path is None:
        parser_type = ModelType.ML_REGRESSION

    if visual_json_path is not None and visual_csv_path is None:
        parser_type = ModelType.ML_CLASSIFICATION

    # evaluation
    if LcConst.RUN_MODEL_EVL_JSON_PATH in run_result_path:
        evl_json_path = run_result_path[LcConst.RUN_MODEL_EVL_JSON_PATH]

    if evl_json_path is not None and run_info_json[LcConst.RUN_OBJ_MODELTYPE] == "DL_CLASSIFICATION":
        parser_type = ModelType.DL_CLASSIFICATION_evl
        print("DL_CLASSIFICATION 평가")
    elif evl_json_path is not None and run_info_json[LcConst.RUN_OBJ_MODELTYPE] == "DL_REGRESSION":
        parser_type = ModelType.DL_REGRESSION_evl
        print("DL_REGRESSION 평가")
    elif evl_json_path is not None and run_info_json[LcConst.RUN_OBJ_MODELTYPE] == "ML_CLASSIFICATION":
        parser_type = ModelType.ML_CLASSIFICATION_evl
        print("ML_CLASSIFICATION 평가")
    elif evl_json_path is not None and run_info_json[LcConst.RUN_OBJ_MODELTYPE] == "ML_REGRESSION":
        parser_type = ModelType.ML_REGRESSION_evl
        print("ML_REGRESSION 평가")
    return parser_type


from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score, mean_absolute_percentage_error
import numpy as np

from Accuinsight.modeler.core.get_for_visual import y_pred_reshape, y_test_reshape


def calculate_regression_metrics(x_val, y_true, model=None):
    if model is not None:
        y_pred = model.predict(x_val)
    else:
        y_pred = x_val

    metrics = {
        'mae': mean_absolute_error(y_true, y_pred),
        'mse': mean_squared_error(y_true, y_pred),
        'rmse': np.sqrt(mean_squared_error(y_true, y_pred)),
        'r2score': r2_score(y_true, y_pred),
        'mape': mean_absolute_percentage_error(y_true, y_pred),
    }

    return metrics


def calculate_classification_metrics(x_val, y_true, average=None, model=None):
    if model is not None:
        y_pred = y_pred_reshape(x_val, model)
    else:
        y_pred = y_pred_reshape(x_val)
    y_true = y_test_reshape(y_true)

    print("y_true shape:", y_true.shape)
    print("y_pred shape:", y_pred.shape)
    result = {
        "y_true": y_true,
        "y_pred": y_pred
    }

    print(result)
    if len(np.unique(y_true)) > 2:
        # 다중 클래스 분류
        print("다중 클래스")
        accuracy = accuracy_score(y_true, y_pred)
        precision = precision_score(y_true, y_pred, average=average)
        recall = recall_score(y_true, y_pred, average=average)
        f1score = f1_score(y_true, y_pred, average=average)
    else:
        # 이진 분류
        print("이진 클래스")
        accuracy = accuracy_score(y_true, y_pred)
        precision = precision_score(y_true, y_pred, average=average, pos_label=1)
        recall = recall_score(y_true, y_pred, average=average, pos_label=1)
        f1score = f1_score(y_true, y_pred, average=average, pos_label=1)

    metrics = {
        'accuracy': accuracy,
        'precision': precision,
        'recall': recall,
        'f1score': f1score
    }

    return metrics


import json
from Accuinsight.modeler.utils.os_getenv import get_os_env
from Accuinsight.modeler.clients.modeler_api_common import WorkspaceRestApi
from Accuinsight.modeler.core.LcConst import LcConst
from collections import OrderedDict


class PackageList:
    def __init__(self):
        self.packageList = OrderedDict()

    def call_api(self, param):
        env_value = get_os_env('ENV')
        modeler_rest = WorkspaceRestApi(env_value[LcConst.BACK_END_API_URL],
                                        env_value[LcConst.BACK_END_API_PORT],
                                        env_value[LcConst.BACK_END_API_URI])

        return modeler_rest.call_rest_api(param, 'packageList')

    def process_response(self, response):
        # JSON 파싱 및 데이터 처리
        try:
            parsed = json.loads(response)
            # print(json.dumps(parsed, indent=4, sort_keys=True)) # 응답 확인

            for item in parsed['data']['list']:
                self.packageList[item['name']] = item['version']
        except json.JSONDecodeError:
            print("JSON 파싱 오류가 발생했습니다.")

    def getCustomPackageList(self, packageType="pip"):
        # 유효한 packageType 검사
        if packageType not in ['all', 'pip', 'conda']:
            raise ValueError("packageType은 'all', 'pip', 'conda' 중 하나여야 합니다.")

        if packageType == 'all':
            # pip 패키지 리스트 처리
            self.packageList = OrderedDict()  # 패키지 리스트 초기화
            response = self.call_api('pip')
            self.process_response(response)
            print("\nAdded/Modified Pip Packages\n")
            for package, version in self.packageList.items():
                print(f'{package}=={version}')
            print("============================================================")
            # conda 패키지 리스트 처리
            self.packageList = OrderedDict()  # 패키지 리스트 초기화
            response = self.call_api('conda')
            self.process_response(response)
            print("\nAdded/Modified Conda Packages\n")
            for package, version in self.packageList.items():
                print(f'{package}=={version}')
        else:
            # 특정 타입의 패키지 리스트 처리
            self.packageList = OrderedDict()  # 패키지 리스트 초기화
            response = self.call_api(packageType)
            self.process_response(response)
            print("\nAdded/Modified " + packageType + " Packages\n\n")
            for package, version in self.packageList.items():
                print(f'{package}=={version}')

        # 안내는 ai runtime 통해서 모델 서빙 하려면 경로의 파일 넣어야함
        print('''
        \n============================================================
airuntime을 통한 모델 배포를 위해 아래 경로에 requirements 파일 추가\n
@주의: pytorch, sklearn, tensorflow 모델에 따라 파일 위치가 다릅니다.

<경로> 
/airuntime/{pytorch|sklearn|tensorflow}/requirements.txt --> 해당 파일(없으면 생성)에 아래 형태로 파일 작성
 
패키지==버전
패키지==버전
패키지==버전
...
...
...
패키지==버전
''')


        # print('''\n============================================================
        # 배포를 위한 requirements 파일 수정\n
        # <pip packages>
        # /airuntime/{pytorch|sklearn|tensorflow}/requirements.txt
        # 해당경로의 파일 수정\n
        # <conda packages>
        # /airuntime/{pytorch|sklearn|tensorflow}/requirements_conda.txt
        # 해당경로의 파일 수정\n
        # ''')